#include <vector>
#include <iostream>
#include <cstdlib>
#include <cctype>

using std::vector;
using std::exit;
using std::cout;
using std::cerr;
using std::endl;

#include "actmap.hh"

static void AgentChan(ActMapBase *map, const string &ag, const string &ch)
{
    unsigned q = ag.find(':');
    unsigned load = 0, rely = 0;
    if(q != ag.npos)
    {
        unsigned a;
        for(a=q; ++a < ag.size() && isdigit(ag[a]); )
            load = load*10 + (ag[a]-'0');
        if(a < ag.size() && ag[a]==':')
        {
            while(++a < ag.size() && isdigit(ag[a]))
                rely = rely*10 + (ag[a]-'0');
        }
    }
    Agent tmp(ag.substr(0, q), load, rely);

    map->AddAgent(tmp);
    map->AddChannel(ch);
    map->SetAgentChan(tmp, ch);
}

#include "mode.hh"

/* parses "dest[,dest[...]] mode [parm[ parm[...]]]" */
static void DoMode(ActMap<Mode> &map, const string &str, bool first)
{
    vector<Channel> chlist;
    
    unsigned a, b=str.size(), c;
    for(a=0; a<b; ++a)
    {
        c=a;
        while(a<b && !isspace(str[a]) && str[a]!=',')++a;
        
        string s = str.substr(c, a-c);
        if(!first)chlist.push_back(s);
        map.AddChannel(s);
        
        if(a >= b || isspace(str[a]))break;
    }
    while(a<b && str[a]==' ')++a;
    for(c=a; a<b && !isspace(str[a]); ++a);
    
    vector<string> parms;
    string mstr = str.substr(c, a-c);
    
    while(a<b && isspace(str[a]))++a;
    for(; a<b; ++a)
    {
        for(c=a; a<b && !isspace(str[a]); ++a);
        string ss = str.substr(c, a-c);
        parms.push_back(ss);
    }
    bool nega = false;
    unsigned p=0;
    for(a=0, b=mstr.size(); a<b; ++a)
    {
        if(mstr[a]=='+') { nega=false; continue; }
        if(mstr[a]=='-') { nega=true; continue; }
        Mode tmp(nega, mstr[a]);
        if(tmp.needsparm())
        {
            if(p >= parms.size())
                continue;
            tmp.setparm(parms[p++]);
        }

        if(first)
            map.AddOp(tmp);
        else
        {
            for(unsigned d=0; d<chlist.size(); ++d)
                map.SetChanOp(chlist[d], tmp);
        }
    }
}

#include "plain.hh"

/* parses "dest[,dest[...]] text" */
static void DoMsg(ActMap<Plain> &map, char type, const string &str, bool first)
{
    vector<Channel> chlist;
    
    unsigned a, b=str.size(), c;
    for(a=0; a<b; ++a)
    {
        c=a;
        while(a<b && !isspace(str[a]) && str[a]!=',')++a;
        
        string s = str.substr(c, a-c);
        if(!first)chlist.push_back(s);
        map.AddChannel(s);
        
        if(a >= b || isspace(str[a]))break;
    }
    while(a<b && str[a]==' ')++a;
    
    Plain tmp(type, str.substr(a));
    if(first)
        map.AddOp(tmp);
    else
    {
        for(unsigned d=0; d<chlist.size(); ++d)
            map.SetChanOp(chlist[d], tmp);
    }
}

#include "kick.hh"

/* parses "dest[,dest[...]] nick[,nick[...]] text" */
static void DoKick(ActMap<Kick> &map, const string &str, bool first)
{
    vector<Channel> chlist;
    
    unsigned a, b=str.size(), c;
    for(a=0; a<b; ++a)
    {
        c=a;
        while(a<b && !isspace(str[a]) && str[a]!=',')++a;
        
        string s = str.substr(c, a-c);
        if(!first)chlist.push_back(s);
        map.AddChannel(s);
        
        if(a >= b || isspace(str[a]))break;
    }
    while(a<b && str[a]==' ')++a;
    
    unsigned sppos = str.find(' ', a);
    
    if(sppos != str.npos) while(sppos < str.size() && str[sppos]==' ')++sppos;
    
    string reason = str.substr(sppos);
    
    for(;;)
    {
        unsigned comma = str.find(',', a);
        if(comma == str.npos || comma > sppos)
            comma = sppos-1;
        
        string nick = str.substr(a, comma-a);
        
        Kick tmp(nick, reason);
        
        if(first)
            map.AddOp(tmp);
        else
        {
            for(unsigned d=0; d<chlist.size(); ++d)
                map.SetChanOp(chlist[d], tmp);
        }
        
        if(comma+1 >= sppos)break;
        
        a = comma+1;
    }
}

static void Usage()
{
    cout <<
        "IRC Action combiner v"VERSION" for agent system like in BisqBot\n"
        "Copyright (C) 1992,2002 Bisqwit (http://iki.fi/bisqwit/)\n"
        "No warranty whatsoever. Distributed under the terms of GPL v2, see COPYING\n"
        "\n"
        "Usage: modecombiner [--notest] <agentdef> [<...>] /<op> [<opdef> [<...>]]\n"
        "\n"
        "\t--notest disables combining parameter limits.\n"
        "\tagentdef is: <nick>[:<load>[:<reliability>]] <channellist>\n"
        "\tchannellist is: \"<chan> [<...>]\"\n"
        "\tFor op='M', opdef is: <modestring>\n"
        "\tFor op='K', opdef is: <kickstring>\n"
        "\tFor op='T', opdef is: <topicstring>\n"
        "\tFor op='G', opdef is: <msgstring>\n"
        "\tFor op='N', opdef is: <noticestring>\n"
        "\tmodestring is: \"<chans> <modes> [<parameter> [<...>]]\"\n"
        "\tkickstring is: \"<chans> <nicks> <text>\"\n"
        "\ttopicstring, msgstring and noticestring are: \"<chans> <text>\"\n"
        "\tchans is: <chan>[,<chan>[<...>]]\n"
        "\tnicks is: <nick>[,<nick>[<...>]]\n"
        "\tmodes is a combination of alphabets and +/- signs.\n"
        "\tload is an unsigned integer, describing the load the agent has.\n"
        "\treliability is an unsigned integer, bigger value meaning more reliable.\n"
        "\t\treliability value is used if two agents have equal chance otherwise.\n"
        "Output consists of lines of following format:\n"
        "\t<op> <nick> <command>\n"
        "\top=W is a special case: it means there is no agent\n"
        "\table to do the job for the particular channel.\n"
        "\n"
    ;
    exit(0);
}
/* Example: modecombiner 
      Bisqwit:0 "#chan #pruu"
      Dih:1 "#blaa #bluu"
      /M
      "mode1"
      "mode2"
*/

static ActMapBase *map;static char opmode;

enum phase_t { getop, analyze, process };

bool TestGroupCount = true;

static void ParseArgs(int argc, const char *const *argv, phase_t phase)
{
    int v;
    for(v=1; v<argc; ++v)
    {
        if(!strcmp(argv[v], "--help"))Usage();
        if(!strcmp(argv[v], "--notest")) { TestGroupCount = false; continue; }
        
        if(argv[v][0] == '/')
        {
            if(phase == getop)
                switch(opmode = argv[v][1])
                {
                    case 'M':
                        map = new ActMap<Mode>;
                        break;
                    case 'K':
                        map = new ActMap<Kick>;
                        break;
                    case 'T':
                    case 'N':
                    case 'G':
                        map = new ActMap<Plain>;
                        break;
                }
            if(!map)
            {
                cerr << "Syntax error\n";
                Usage();
            }
            ++v;
            break;
        }
        
        string agnam = argv[v];
        string s = argv[++v];
        
        if(phase == analyze)
        {
            for(unsigned a=0, b=s.size(); a<b; ++a)
            {
                unsigned c = a;
                while(a<b && s[a]!=' ' && s[a]!=',')++a;
                
                if(a > c)
                {
                    string ss = s.substr(c, a-c);
                    AgentChan(map, agnam, ss);
                }
            }
            /* For umodes: */
            if(opmode == 'M')
                AgentChan(map, agnam, agnam);
        }
    }
    if(v == argc)
    {
        Usage();
    }
    while(v < argc)
    {
        if(phase == analyze || phase == process)
        {
            switch(opmode)
            {
                case 'M':
                    DoMode(* (ActMap<Mode> *) (map), argv[v], (phase == analyze));
                    break;
                case 'K':
                    DoKick(* (ActMap<Kick> *) (map), argv[v], (phase == analyze));
                    break;
                case 'T':
                case 'N':
                case 'G':
                    DoMsg(* (ActMap<Plain> *) (map), opmode, argv[v], (phase == analyze));
                    break;
            }
        }
        ++v;
    }
}

int main(int argc, const char *const *argv)
{
    /* See what will we do. */
    ParseArgs(argc, argv, getop);
    
    /* List all channels and agents and modes. */
    ParseArgs(argc, argv, analyze);
    
    map->InitMap();
    
    /* Now build the map. */
    ParseArgs(argc, argv, process);
    
    map->Distribute();

    return 0;
}
